#!/bin/bash

# mlv2badpixels.sh
#
# http://www.magiclantern.fm/forum/index.php?topic=16054
#
# Create a dcraw ".badpixels" formatted file from
# an MLV file (Magic Lantern raw video)
# For removing focus pixels from Magic Lantern raw video
# files shot on on Canon 100D, 650D, 700D and EOSM cameras.
#
# The focus pixel map files (.fpm) for the raw video buffer
# need to be in the same working path as this script.
#
# Mode          Buffer Size   Notes
# mv1080      - 1808x1190  -  not available on the EOSM.
# mv720       - 1808x727   -  needs to be stretched by 1.67x
#                             vertically - up to 50/60 fps
#                             except on EOSM.
# mv640       - 1808x1190  -  basically the same as mv1080 mode.
# mv1080crop  - 1872x1060  -  highest resolution and largest
#                             sustainable frame sizes on all cameras.
# mv640crop   -            -  not possible.
# zoom        - 2592x1108  -  same 5x magnification as mv1080crop but
#                             can pan around sensor and can achieve
#                             higher resolution than mv1080crop.
#
# Sorry but with zoom mode a new badpixels file needs to be made
# every time that the pan position changes.
#
# This script does not support filenames and paths with spaces
# and there is very little error checking.
#
# Copyright (C) 2016 Magic Lantern Team
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor,
# Boston, MA  02110-1301, USA.
#/

usage()
{
cat << EOF
Usage: mlv2badpixels.sh [-hmv][-o output_file] <input_file>
Parameters:
    -h           Show this message
    -o filename  Customize Filename
    -v           verbose

example:
    ./mlv2badpixels.sh -v -o M21-1747_1_2015-12-06_0001_C0000.txt M21-1747.MLV

    creates a file named M21-1747_1_2015-12-06_0001_C0000.txt that can be
    used to remove the focus pixels on the dng files extracted from M21-1747.MLV

This script will create a dcraw "badpixels" file
to remove focus pixels from an MLV file shot with
cameras that show their focus pixels in raw video.

EOS Rebel T4i / 650D / Kiss X6i
EOS Rebel T5i / 700D / Kiss X7i
EOS Rebel SL1 / 100D / Kiss X7
EOS M

EOF
}

VERBOSE=
output=

workingDir=`dirname "$0"`
cd "${workingDir}"

while getopts “hmo:v” OPTION; do
  case $OPTION in
  h)usage
    exit 0
    ;;
  o)output="$OPTARG"
    ;;
  v)VERBOSE=1
    ;;
  :)echo "Error: -$OPTARG requires an argument"
    usage
    exit 1
    ;;
  ?)usage
    exit 1
    ;;
  esac
done

shift $((OPTIND - 1))

if [ $VERBOSE ]; then echo workingDir = $workingDir; fi

MLV="$1"

if [[ -z $MLV ]]; then
  usage
  exit 1
fi

if [ $VERBOSE ]; then echo input file = $MLV; fi

camera_model=`mlv_dump -m -v "$MLV" | grep Camera"[[:space:]]"Model | head -1 | sed -e 's/^[ Camera \s Model: \s 0x]*//'`

resolution=`mlv_dump -m -v "$MLV" | grep Res | head -1 | sed -e 's/^[ Res:]*//'`
pan=`mlv_dump -m -v "$MLV" | grep Pan | head -1 | sed -e 's/^[ Pan:]*//'`

IFS='x'
  read -r width height <<< "$resolution"
  read -r panPosX panPosY <<< "$pan"
unset IFS

cropX=$((panPosX + 7 & ~7))
cropY=$((panPosY & ~1))

if [ $VERBOSE ]; then echo Resolution = $resolution; echo Pan = $pan; echo Crop = $cropX"x"$cropY; fi

raw_width=`mlv_dump -m -v "$MLV" | grep width | head -1 | sed -e 's/^[ width]*//'`
raw_height=`mlv_dump -m -v "$MLV" | grep height | head -1 | sed -e 's/^[ height]*//'`
raw_buffer=$raw_width"x"$raw_height

#Heuretic code. If close to match fpm file straighten things right
if [[ "$raw_buffer" = \1808x11* ]]; then
raw_buffer=1808x1190
elif
[[ "$raw_buffer" = \1808x72* ]]; then
raw_buffer=1808x727
elif
[[ "$raw_buffer" = \1872x10* ]]; then
raw_buffer=1872x1060
elif
[[ "$raw_buffer" = \2592x110* ]]; then
raw_buffer=2592x1108
fi

fpm=$camera_model"_"$raw_buffer".fpm"

if [ $VERBOSE ]; then echo Camera Model = $camera_model; fi

case $camera_model in
  80000301)
    camera=650D
    ;;
  80000326)
    camera=700D
    ;;
  80000331)
    camera=EOSM
    ;;
  80000346)
    camera=100D
    ;;
  *) exit 2
    ;;
esac

if [ $VERBOSE ]; then
  echo Camera = $camera
  echo Raw Buffer = $raw_buffer
  echo Map File = $fpm
fi

case $raw_buffer in
  1808x11**)
    video_mode=mv1080
    ;;
  1808x72*)
    video_mode=mv720
    ;;
  1872x10**)
    video_mode=mv1080crop
    ;;
  2592x110*)
    video_mode=zoom
    ;;
  *) exit 3
    ;;
esac

if [ $VERBOSE ]; then echo Video Mode = $video_mode; fi

if [ -z "$output" ]; then
    output=$camera"_"$video_mode"_"$resolution".txt"
fi

if [ $VERBOSE ]; then echo Output File = "$output"; fi

if test -f "$output"; then
  rm "$output"
  if [ $VERBOSE ]; then echo Removing previous file = "$output"; fi
fi

while read -r raw_buffer_x raw_buffer_y; do
  ((x = $raw_buffer_x - $cropX))
  ((y = $raw_buffer_y - $cropY))
  if (( "$x" >= 0 )) && (( "$y" >= 0 )); then
    if (( "$x" <= "$width" )) && (( "$y" <= "$height" )); then
      echo -e "$x \t $y \t 0" >> "$output"
    fi
  fi
done < "$fpm"
